package tech.nevets.vcardgen;

import com.google.gson.Gson;
import com.google.gson.JsonObject;
import spark.Request;
import spark.Response;
import spark.Route;

import javax.imageio.*;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageOutputStream;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

public class VCard {
    private static final Color WHITE = new Color(255, 255, 255);
    private static final Color GREEN = new Color(101, 142, 61);

    private static final Font ARIAL65 = new Font("Arial", Font.PLAIN, 65);
    private static final Font ARIAL55 = new Font("Arial", Font.PLAIN, 55);
    private static final Font ARIAL45 = new Font("Arial", Font.PLAIN, 45);
    private static final Font ARIAL45I = new Font("Arial", Font.ITALIC, 45);
    private static final Font ARIAL44I = new Font("Arial", Font.ITALIC, 44);
    private static final Font ARIAL40I = new Font("Arial", Font.ITALIC, 40);
    private static final Font ARIAL38I = new Font("Arial", Font.ITALIC, 38);

    private static final String[] KEYS = {"Name", "Title", "Email", "Location", "Address", "SchoolNumber", "Extension", "DirectNumber", "CellNumber"};

    private BufferedImage rawImage = new BufferedImage(1080, 602, BufferedImage.TYPE_INT_ARGB);
    private BufferedImage workingImage = new BufferedImage(1080, 602, BufferedImage.TYPE_INT_ARGB);
    private IIOImage finalImage;

    private final String[] data = new String[9];
    private final boolean[] flags = new boolean[3]; // 0 - Long Address, 1 - Has Direct Number, 2 - Has Cell Phone
    private final BufferedImage background;
    private final int size;

    public VCard(JsonObject json) {
        this.data[0] = json.get("name").getAsString(); //name
        this.data[1] = json.get("title").getAsString(); //title
        this.data[2] = json.get("email").getAsString(); //email
        Location location = Location.getLocation(json.get("locationId").getAsString());
        this.data[3] = location.getName(); //locationName
        this.data[4] = location.getAddress(); //address
        this.data[5] = location.getNumber(); //schoolNumber
        background = location.getBackground();
        this.data[6] = json.get("extension").getAsString(); //extension
        this.data[7] = json.get("directNumber").getAsString(); //directNumber
        this.data[8] = json.get("cellNumber").getAsString(); //cellNumber
        this.flags[1] = this.data[7].length() > 0; //hasDirectNumber
        this.flags[2] = this.data[8].length() > 0; //hasCellNumber
        size = json.get("size").getAsInt();

        renderImage();
        resizeImage();
        try {
            addMetadata();
        } catch (IOException e) {
            System.out.println("Error adding metadata");
            e.printStackTrace();
        }
    }

    public VCard(String name, String title, String email, String locationId, String extension, String directNumber, String cellNumber, int size) {
        this.data[0] = name; //name
        this.data[1] = title; //title
        this.data[2] = email; //email
        Location location = Location.getLocation(locationId);
        this.data[3] = location.getName(); //locationName
        this.data[4] = location.getAddress(); //address
        this.data[5] = location.getName(); //schoolNumber
        background = location.getBackground();
        this.data[6] = extension; //extension
        this.data[7] = directNumber; //directNumber
        this.data[8] = cellNumber; //cellNumber
        this.flags[1] = this.data[7].length() > 0; //hasDirectNumber
        this.flags[2] = this.data[8].length() > 0; //hasCellNumber
        this.size = size;

        renderImage();
        resizeImage();
        try {
            addMetadata();
        } catch (IOException e) {
            System.out.println("Error adding metadata");
            e.printStackTrace();
        }
    }

    private void renderImage() {
        Graphics2D g = workingImage.createGraphics();
        g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

        g.drawImage(background, 0, 0, null);

        g.setColor(WHITE);
        if (fitsDimensions(data[0], ARIAL65, 969)) g.setFont(ARIAL65);
        else g.setFont(ARIAL55);
        g.drawString(data[0], 85, 112);

        if (fitsDimensions(data[1], ARIAL45I, 970)) g.setFont(ARIAL45I);
        else g.setFont(ARIAL40I);
        g.drawString(data[1], 89, 176);

        int doubleNumOffset = 0;
        if (flags[2]) {
            doubleNumOffset = 40;
        }

        int longAddrOffset = fitsDimensions(data[4], ARIAL38I, 694) ? 0 : 40;

        g.setFont(ARIAL45);
        g.drawString(data[2], 62, 380 - doubleNumOffset - longAddrOffset);

        g.setColor(GREEN);
        g.setFont(ARIAL44I);
        g.drawString(data[3], 59, 447 - doubleNumOffset - longAddrOffset);

        if (!fitsDimensions(data[4], ARIAL38I, 694)) {
            StringBuilder addrLineOne = new StringBuilder();
            StringBuilder addrLineTwo = new StringBuilder();

            String[] splitAddr = data[4].split(",");
            int i = 0;
            for (int width = 0; width < 694; i++) {
                int splitSize = g.getFontMetrics().stringWidth(splitAddr[i]);
                if ((width + splitSize) < 694) {
                    width += splitSize;
                    addrLineOne.append(splitAddr[i]);
                    addrLineOne.append(",");
                } else {
                    addrLineOne.deleteCharAt(addrLineOne.length() - 1);
                    break;
                }
            }

            for (; i < splitAddr.length; i++) {
                addrLineTwo.append(splitAddr[i]);
                addrLineTwo.append(",");
            }
            addrLineTwo.deleteCharAt(addrLineTwo.length() - 1);

            g.setFont(ARIAL38I);
            g.drawString(addrLineOne.toString().trim(), 59, 491 - doubleNumOffset - longAddrOffset);
            g.drawString(addrLineTwo.toString().trim(), 59, 491 - doubleNumOffset);
        } else {
            g.drawString(data[4], 59, 491 - doubleNumOffset);
        }

        String number;
        int numY = flags[2] ? 496 : 540;
        if (!flags[1]) number = "W: " + data[5] + " x" + data[6];
        else number = "W: " + data[7];
        g.drawString(number, 59, numY);
        if (flags[2]) g.drawString("C: " + data[8], 59, 540);

        g.dispose();
        rawImage = workingImage;
    }

    public void resizeImage() {
        resizeImage(this.size);
    }

    public void resizeImage(int size) {
        BufferedImage resizedImage;
        switch (size) {
            case 0 -> resizedImage = resizeImage(300, 167, false);
            case 1 -> resizedImage = rawImage;
            default -> resizedImage = background;
        }
        workingImage = resizedImage;
    }

    public BufferedImage resizeImage(int width, int height, boolean antiAlias) {
        BufferedImage resizedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = resizedImage.createGraphics();
        g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        if (antiAlias) g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        g.drawImage(rawImage, 0, 0, width, height, null);
        g.dispose();
        return resizedImage;
    }

    public void addMetadata() throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ImageIO.write(workingImage, "png", baos);

        ImageReader reader = ImageIO.getImageReadersByMIMEType("image/png").next();
        reader.setInput(ImageIO.createImageInputStream(new ByteArrayInputStream(baos.toByteArray())));
        finalImage = reader.readAll(0, null);

        IIOMetadataNode text = new IIOMetadataNode("tEXt");
        for (int i = 0; i < KEYS.length; i++) {
            IIOMetadataNode textEntry = new IIOMetadataNode("tEXtEntry");
            textEntry.setAttribute("keyword", KEYS[i]);
            textEntry.setAttribute("value", data[i]);
            text.appendChild(textEntry);
        }
        IIOMetadataNode root = new IIOMetadataNode("javax_imageio_png_1.0");
        root.appendChild(text);

        finalImage.getMetadata().mergeTree("javax_imageio_png_1.0", root);
    }

    public byte[] toByteArray() throws IOException {
        ImageWriter writer = ImageIO.getImageWritersByMIMEType("image/png").next();
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ImageOutputStream ios = ImageIO.createImageOutputStream(baos);
        writer.setOutput(ios);
        writer.write(finalImage);
        return baos.toByteArray();
    }

    private static final Canvas C = new Canvas();
    private static boolean fitsDimensions(String text, Font font, int maxLength) {
        return C.getFontMetrics(font).stringWidth(text) < maxLength;
    }

    public static String getDataFromVCard(InputStream rawImage) throws IOException {
        ImageReader reader = ImageIO.getImageReadersByMIMEType("image/png").next();
        reader.setInput(ImageIO.createImageInputStream(rawImage));

        IIOMetadata metadata = reader.getImageMetadata(0);
        IIOMetadataNode root = (IIOMetadataNode) metadata.getAsTree("javax_imageio_png_1.0");
        IIOMetadataNode text = (IIOMetadataNode) root.getElementsByTagName("tEXt").item(0);
        int numTextEntries = text.getLength();

        Map<String, String> keyValueMap = new HashMap<>();
        for (int i = 0; i < numTextEntries; i++) {
            IIOMetadataNode textEntry = (IIOMetadataNode) text.item(i);
            String key = textEntry.getAttribute("keyword");
            String value = textEntry.getAttribute("value");
            if (key.equals("Address") || key.equals("SchoolNumber")) {
                continue;
            }
            if (key.equals("Location")) {
                value = Location.getLocationId(value);
            }
            keyValueMap.put(key, value);
        }

        return new Gson().toJson(keyValueMap);
    }
}
